import com.dearle.groovydsl.java.*
import com.dearle.groovydsl.test.SpockScriptSpecification

class ChapterFiveSpec extends SpockScriptSpecification  {

    def "we can lose a reference to a Closure" () {
        given: "we have a script to run"
            def script = new File("scripts/ChapterFive/LostClosure.groovy")

        when: "we run the script"
            shell.evaluate script

        then:
            """Hello, Fred
Hello, Barney
Hello, Wilma""" == output()
    }

    def "Closures are not executed until we call() them explicitly" () {
        given: "we have a script to run"
            def script = new File("scripts/ChapterFive/ClosureOrder.groovy")

        when: "we run the script"
            shell.evaluate script

        then:
            """one
three
two
four""" == output()
    }

    def "we can iterate an object with the each method"() {
        given: "an Integer"
            def number = 1
        when: "we call the each method on it"
            number.each { println it }
        then: "just the object itself gets passed into the Closure"
            "1" == output()
    }

    def "we can iterate a String with the each method"() {
        given: "a String"
            def string = "String"
        when: "we call the each method on the String"
            string.each { println it }
        then: "each knows to iterate the chars in the String"
            """S
t
r
i
n
g""" == output()
    }

    def "calling a method which accepts multiple params and a Closure"() {
        given:
            def flintstones = ["Fred", "Barney", "Wilma"]
        when: "we call findIndexOf passing int and a Closure"
            def result = flintstones.findIndexOf(0) { it == 'Wilma'}
        then:
            result == 2
    }

    def "can invoke method which accepts a single Closure parameter"() {
        when: "we invoke a method that accepts a closure"
            closureMethod {
                println "Closure called"
            }
        then: "the Closure passed in was executed"
            "Closure called" == output()
    }

    def "can invoke method which accepts a Closure with other parameters"() {
        when: "we invoke a method that accepts a String and a Closure"
            closureMethodInteger(1) {
                println "Line 2"
            }
        then: "the Closure passed in was executed with the parameter"
            """Line 1
Line 2""" == output()
    }

    def "it can be useful to pass parameters on to the Closure"() {
        when: "we invoke a method that accepts a String and a Closure"
            closureMethodString("Dolly") { name ->
                println "Hello, $name"
            }
        then: "the Closure passed in was executed with the parameter"
            """Greet someone
Hello, Dolly""" == output()
    }

    def "we can write a mini DSL with just closures" () {
         given: "we have a script to run"
             def script = new File("scripts/ChapterFive/TransactionLock.groovy")

         when: "we run the script"
             shell.evaluate script

         then:
             """Transaction lock
Customer name was Fred
Customer name is now Barney
Transaction release""" == output()
     }

    def "closure can have zero to many typed or dynamic parameters"() {
        given: "Closures with various parameter definitione"
            def defaultParams = { println it; }
            def dynamicParams = { something -> println something; }
            def intParams = { int something -> println something; }
            def stringParams = { String something -> println something; }
            def noParams = { -> }
        when: "Invoking these with valid parameters"
            defaultParams 1
            defaultParams "String"
            dynamicParams 1
            dynamicParams "String"
            intParams 1
            stringParams "String"
            noParams ()
        then: "we expect them to work"
            notThrown Exception
        when: "we pass an incorrect type for a typed parameter"
            stringParams 1
        then: "we expect that Groovy won't find a matching Closure"
            thrown MissingMethodException
        when: "we pass too many parameters"
            dynamicParams "String1", 1
        then: "that should fail also"
            thrown MissingMethodException
        when: "passing a parameter to a Closure that does not expect one"
            noParams "String"
        then: "we expect that to fail"
            thrown MissingMethodException
    }

    def "can implement a Closure in pure Java"() {
        given: "an instance of a Java Closure"
            def stringParams = new StringClosure(this)
        when: "Invoking this with the parameters defined by the doCall"
            stringParams "String"
        then: "we expect them to work"
            notThrown Exception
        when: "we pass an incorrect type for the doCall parameter"
            stringParams 1
        then: "we expect that Groovy won't find a matching Closure"
            thrown MissingMethodException
    }

    def "can implement multiple doCalls for a Closure in pure Java"() {
        given: "an instance of a Java Closure"
            def multiParams = new MultiClosure(this)
        when: "Invoking these with the parameters defined by the doCall"
            multiParams "String"
        then: "we expect them to work"
            notThrown Exception
        when: "Invoking these with the parameters defined by the doCall"
            multiParams 1
        then: "we expect them to work"
            notThrown Exception
    }

    def closureMethod(Closure c) {
        c.call()
    }

    def closureMethodInteger(Integer i, Closure c) {
        println "Line $i"
        c.call()
    }

    def closureMethodString(String s, Closure c) {
        println "Greet someone"
        c.call(s)
    }

}
